package com.kedda.carApp

import android.annotation.SuppressLint
import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothDevice
import android.content.Intent
import android.os.Bundle
import android.os.Handler
import android.os.Looper
import android.os.SystemClock
import android.util.Log
import android.view.MotionEvent
import android.widget.Button
import android.widget.Switch
import android.widget.Toast
import androidx.appcompat.widget.AppCompatImageButton
import androidx.lifecycle.lifecycleScope
import kotlinx.coroutines.launch
import java.text.SimpleDateFormat
import java.util.Date
import java.util.Locale
import java.util.UUID

class ControlActivity : MainActivity() {

    companion object {
        var m_myUUID: UUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB")
        lateinit var m_bluetoothAdapter: BluetoothAdapter
        lateinit var m_address: String
    }

    private var commandHandler: Handler? = null
    private var commandRunnable: Runnable? = null

    // Time formatter for precise timestamps
    private val timeFormatter = SimpleDateFormat("HH:mm:ss.SSS", Locale.getDefault())

    @SuppressLint("UseSwitchCompatOrMaterialCode", "WrongViewCast", "MissingInflatedId")
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.control_layout)

        m_address = intent.getStringExtra(EXTRA_ADDRESS).toString()
        m_bluetoothAdapter = BluetoothAdapter.getDefaultAdapter()

        val device: BluetoothDevice? = m_bluetoothAdapter.getRemoteDevice(m_address)

        // Move Bluetooth connection to a coroutine
        if (device != null) {
            lifecycleScope.launch {
                val success = BluetoothManager.connectToDevice(this@ControlActivity, device, m_myUUID)
                if (!success) {
                    logWithTime("Connection failed in ControlActivity", isError = true)
                    val intent = Intent(this@ControlActivity, MainActivity::class.java)
                    intent.putExtra(EXTRA_ADDRESS, m_address)
                    startActivity(intent)
                    finish()
                }
            }
        } else {
            logWithTime("Bluetooth device not found", isError = true)
            Toast.makeText(this, "Bluetooth Device Not Found", Toast.LENGTH_SHORT).show()
        }

        val switchLayout = findViewById<Switch>(R.id.switch1)
        switchLayout.isChecked = false

        switchLayout.setOnCheckedChangeListener { _, isChecked ->
            if (isChecked) {
                logWithTime("Voice mode switch activated")
                lifecycleScope.launch {
                    BluetoothManager.sendCommand("VoiceModeAccessed\n")
                }
                val intent = Intent(this, VoiceControlActivity::class.java)
                intent.putExtra(EXTRA_ADDRESS, m_address)
                startActivity(intent)
                finish()
            } else {
                logWithTime("Button mode switch activated")
                lifecycleScope.launch {
                    BluetoothManager.sendCommand("ButtonModeAccessed\n")
                }
            }
        }

        // Setting button movement commands
        setupMovementButton(findViewById(R.id.forwardButton), "forwards\n")
        setupMovementButton(findViewById(R.id.backwardsdButton), "backwards\n")
        setupMovementButton(findViewById(R.id.leftButton), "left\n")
        setupMovementButton(findViewById(R.id.rightButton), "right\n")
        setupMovementButton(findViewById(R.id.findRocketButton), "FindRocket\n")

        // Setting stop button
        val controlStop = findViewById<AppCompatImageButton>(R.id.stopButton)
        controlStop.setOnClickListener {
            logWithTime("Stop button pressed")
            lifecycleScope.launch {
                BluetoothManager.sendCommand("motorstop\n")
            }
        }

        // setting the gripper buttons
        val grabButton = findViewById<Button>(R.id.graspButton)
        grabButton.setOnClickListener {
            logWithTime("Grab button pressed")
            lifecycleScope.launch {
                BluetoothManager.sendCommand("grabrocket\n")
            }
        }

        val release = findViewById<Button>(R.id.releaseButton)
        release.setOnClickListener {
            logWithTime("Release button pressed")
            lifecycleScope.launch {
                BluetoothManager.sendCommand("releaserocket\n")
            }
        }

        // Setting disconnect button
        val disconnectButton = findViewById<Button>(R.id.disconnectButton)
        disconnectButton.setOnClickListener {
            logWithTime("Disconnect button pressed")
            lifecycleScope.launch {
                BluetoothManager.disconnect()
                val intent = Intent(this@ControlActivity, MainActivity::class.java)
                intent.putExtra(EXTRA_ADDRESS, m_address)
                startActivity(intent)
                logWithTime("Disconnected from device")
                finish()
            }
        }
    }

    // Helper function to log with timestamp
    private fun logWithTime(message: String, isError: Boolean = false) {
        val timestamp = timeFormatter.format(Date())
        val logMessage = "[$timestamp] $message"
        if (isError) {
            Log.e("ControlActivity", logMessage)
        } else {
            Log.d("ControlActivity", logMessage)
        }
    }

    // If button pressed send command continuously while pressed
    @SuppressLint("ClickableViewAccessibility")
    private fun setupMovementButton(button: AppCompatImageButton, command: String) {
        button.setOnTouchListener { _, motionEvent ->
            when (motionEvent.action) {
                MotionEvent.ACTION_DOWN -> {
                    val pressTime = SystemClock.elapsedRealtimeNanos()
                    val pressTimestamp = timeFormatter.format(Date())
                    logWithTime("PRESSED - Button: ${button.contentDescription} - Command: $command - SystemNanos: $pressTime")
                    startSendingCommand(command)
                    true
                }
                MotionEvent.ACTION_UP -> {
                    val releaseTime = SystemClock.elapsedRealtimeNanos()
                    val releaseTimestamp = timeFormatter.format(Date())
                    logWithTime("RELEASED - Button: ${button.contentDescription} - Command: $command - SystemNanos: $releaseTime")
                    stopSendingCommand()
                    true
                }
                else -> false
            }
        }
    }

    // Start sending command every 100ms
    private fun startSendingCommand(command: String) {
        commandHandler = Handler(Looper.getMainLooper())
        commandRunnable = object : Runnable {
            override fun run() {
                lifecycleScope.launch {
                    BluetoothManager.sendCommand(command)
                    logWithTime("SENDING - Command: $command")
                }
                commandHandler?.postDelayed(this, 1000) // Send command every 100ms
            }
        }
        commandHandler?.post(commandRunnable!!)
    }

    // Stop sending command
    private fun stopSendingCommand() {
        logWithTime("STOPPED - Command sending")
        commandHandler?.removeCallbacks(commandRunnable ?: return)
        commandHandler = null
    }
}