/*
 * ESP32 Smart Car Controller
 * Features:
 * - Bluetooth control (via SerialBT)
 * - MQTT connectivity for remote commands
 * - Ultrasonic obstacle detection
 * - Rocket search & grab mechanism with servo
 * - Two control modes: Button and Voice
 */ 
 
 
#include <WiFi.h>  // for WiFi
#include <PubSubClient.h> // for mqtt connection
#include <BluetoothSerial.h> //for bluetooth control
#include <Arduino.h> // Arduino lib
#include <ESP32Servo.h> // for servo motor control 

// ======== Pin Definitions ========
#define SOUND_SPEED 0.034 // FOR ULTRASONIC DISTANCE CALUCLATION
const int IN1 = 12;  // Motor L forward
const int IN2 = 14;  // Motor L back
const int IN3 = 27;  // Motor R forward
const int IN4 = 26;  // Motor R back
const int trigPin = 5;   // ULTRASONIC TRIGGER PIN
const int echoPin = 18;  // ULTRASONIC ECHO PIN 
const int ENA = 33; // Connect to ENA on the L298N
const int ENB = 32; // Connect to ENB on the L298N

const char* ssid = "SSID"; // insert hotspot or wifi ID
const char* password = "PASSWORD"; // insert WiFi password
const char* mqtt_server = "mqtt.eclipseprojects.io"; // MQTT Brocker address
static const int servoPin = 23; // Servo pin

// Global variable to track the time of the last received command
unsigned long lastCommandTime = 0; //Tracks last received command time
const unsigned long COMMAND_TIMEOUT = 1000; // milliseconds

Servo servo1; // servo object
BluetoothSerial SerialBT_App; // Bluetooth interface 
WiFiClient espClient;  // wifi client
PubSubClient client(espClient);  // mqtt client
String deviceName = "ESP32-BT-CARRR"; // bluetooth device name

bool clientConnected = false; // bluetooth connection status
String currentMode = "button"; // default start mode
bool rocketFound = false; // rocket detection flag
bool objectDetected = false; // obstacle detection flag

void callback(char* topic, byte* payload, unsigned int length) {
    // Handles incoming MQTT messages
  Serial.print("Message received on topic: ");
  Serial.println(topic);
  Serial.print("Message: ");
  for (int i = 0; i < length; i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();
}

void setupMotors() {
     // Initialize motor control pins 
    pinMode(IN1, OUTPUT);
    pinMode(IN2, OUTPUT);
    pinMode(IN3, OUTPUT);
    pinMode(IN4, OUTPUT);
    motorStop(); // to make sure motors start in stopped state
}

void motorControl(uint8_t in1, uint8_t in2, uint8_t in3, uint8_t in4) {
  // Controls motor directions using H-bridge
    digitalWrite(IN1, in1);
    digitalWrite(IN2, in2);
    digitalWrite(IN3, in3);
    digitalWrite(IN4, in4);
}

void motorStop() {
  // Stops both motors
    motorControl(LOW, LOW, LOW, LOW);
}
void reconnectWiFi() {
  // Reconnects to WiFi if connection is lost
    if (WiFi.status() != WL_CONNECTED) {
        Serial.println("WiFi lost, reconnecting...");
        WiFi.begin(ssid, password);
        int attempts = 0;
        while (WiFi.status() != WL_CONNECTED && attempts < 10) {
            delay(500);
            Serial.print(".");
            attempts++;
        }
        if (WiFi.status() == WL_CONNECTED) {
            Serial.println("\nWiFi reconnected!");
        } else {
            Serial.println("\nFailed to reconnect to WiFi.");
        }
    }
}

void processCommand(String command) {
  // Main command processor for both modes
    command.trim();
    command.toLowerCase();
    
    // Update last command time whenever a command is received
    lastCommandTime = millis();
    
    // Set the current mode if needed
    if (command == "buttonmodeaccessed" || command == "voicemodeaccessed") {
        currentMode = command;
        return;
    }
    
    // We assume button mode by default
    currentMode = "buttonmodeaccessed";
    // Route command to appropriate handler
    if (currentMode == "buttonmodeaccessed") {
        handleButtonMode(command);
    } else if (currentMode == "voicemodeaccessed") {
        handleVoiceMode(command);
    }
}
    
void handleVoiceMode(String command) {
  // Voice commands with timed movements
    if (command == "forwards") {
        if (getDistanceCM() >= 5) { // Only move if path is clear
            motorControl(LOW, HIGH, LOW, HIGH);
            delay(2000);
            motorStop();
        } else {
            avoidObstacle();
        }
    } else if (command == "backwards") {
        motorControl(HIGH, LOW, HIGH, LOW);
        delay(2000);
        motorStop();
    } else if (command == "left") {
        if (getDistanceCM() >= 5) { // Only move if path is clear
            motorControl(LOW, HIGH, HIGH, LOW);
            delay(2000);
            motorStop();
        } else {
            avoidObstacle();
        }
    } else if (command == "right") {
        if (getDistanceCM() >= 5) {// Only move if path is clear
            motorControl(HIGH, LOW, LOW, HIGH);
            delay(2000);
            motorStop();
        } else {
            avoidObstacle();
        }
    }  else {
        motorStop();  // Default action for unknown commands
    }
}
void handleButtonMode(String command) {
    if (command == "forwards") {
        if (getDistanceCM() >= 5) {// Only move if path is clear
            motorControl(LOW, HIGH, LOW, HIGH);
        } else {
            avoidObstacle();
        }
    } else if (command == "backwards") {
        motorControl(HIGH, LOW, HIGH, LOW);
    } else if (command == "left") {
        if (getDistanceCM() >= 5) {// Only move if path is clear
            motorControl(LOW, HIGH, HIGH, LOW);
        } else {
            avoidObstacle();
        }
    } else if (command == "right") {
        if (getDistanceCM() >= 5) {// Only move if path is clear
            motorControl(HIGH, LOW, LOW, HIGH);
        } else {
            avoidObstacle();
        }
    } else if (command == "grabrocket") {
        servo1.write(180);
        Serial.println("Grabbed rocket.");
    } else if (command == "releaserocket") {
        servo1.write(0);
        Serial.println("Released rocket.");
    } else if (command == "motorstop") {
        servo1.write(180);
        delay(1020);
        motorStop();
    } else if (command == "findrocket") {
        findRocket();  // No obstacle avoidance when searching for the rocket
    } else {
        motorStop(); // Default action for unknown commands
    }
}
bool checkMQTTMessage() {
    client.loop(); 
}
void findRocket() {
    reconnectWiFi();
    rocketFound = false;
    servo1.write(180);
    for (int i = 0; i < 12; i++) {
        if (rocketFound) {
            Serial.println("Rocket found, stopping search!");
            return;
        }
        client.loop();
        if (WiFi.status() != WL_CONNECTED) reconnectWiFi();
        
        if (getDistanceCM() <= 5) {
            Serial.println("Rocket detected! Grabbing...");
            servo1.write(0);
            return;
        }
        motorControl(LOW, HIGH, HIGH, LOW);
        delay(320);
        motorStop();
        delay(5000);
    }
    Serial.println("Rocket not found after full sweep.");
}

void avoidObstacle() {
    //  obstacle avoidance routine

    Serial.println("Obstacle detected, avoiding...");
    motorStop();
    delay(500);
    motorControl(HIGH, LOW, HIGH, LOW);  // Move backward
    delay(500);
    motorStop();
    motorControl(HIGH, LOW, LOW, HIGH);  // Turn right
    delay(500);
    motorStop();
}


float getDistanceCM() {
  // Measures distance using ultrasonic sensor
    digitalWrite(trigPin, LOW);
    delayMicroseconds(2);
    digitalWrite(trigPin, HIGH);
    delayMicroseconds(10);
    digitalWrite(trigPin, LOW);
    long duration = pulseIn(echoPin, HIGH);
    long distance = duration * 0.034 / 2;
    //Serial.println(distance);
    delay(10);
    return distance;
}
void connectWiFi(){


}

void setup() {
  // Hardware initialization

    Serial.begin(115200);
    // Motor setup

    pinMode(ENA, OUTPUT);
    pinMode(ENB, OUTPUT);
    delay(1000);
    analogWrite(ENA, 100);
    analogWrite(ENB, 100);
    // Servo setup
    servo1.attach(servoPin, 500, 2400);
    delay(1000);
    // WiFi connection
    WiFi.begin(ssid, password);
    while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        Serial.print(".");
    }
    // MQTT setup
    Serial.println("\nWiFi connected");
    client.setServer(mqtt_server, 1883);
    client.setCallback(callback);
    while (!client.connected()) {
        if (client.connect("ESP32-CAM-ID1869482003FR")) {
            Serial.println("Connected to MQTT broker");
            client.subscribe("esp32/cam/detection/esp32Car");
        } else {
            Serial.print("Failed, rc=");
            Serial.print(client.state());
            Serial.println(" trying again in 5 seconds");
            delay(2000);
        }
    } 
    // Bluetooth setup
    SerialBT_App.begin(deviceName);
    setupMotors();
    // ultrasonic sensor setup
    pinMode(LED1, OUTPUT);
    pinMode(trigPin, OUTPUT);
    pinMode(echoPin, INPUT);
}

void loop() {
     // Bluetooth connection management
    bool currentConnection = SerialBT_App.hasClient();
    if (currentConnection != clientConnected) {
        clientConnected = currentConnection;
        Serial.println(clientConnected ? "Client connected!" : "Client disconnected.");
    }
    // Process Bluetooth commands
    if (clientConnected && SerialBT_App.available()) {
        String command = SerialBT_App.readStringUntil('\n');
         Serial.println(command);
       
        processCommand(command);
    }
    
    // If no command received within COMMAND_TIMEOUT, stop the motors
    if (millis() - lastCommandTime > COMMAND_TIMEOUT) {
        motorStop();
    }
    // Continuous distance monitoring
    getDistanceCM();
    
    delay(10);
}